﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;
using System.Runtime.InteropServices;
using System.Windows.Forms;

using Microsoft.Win32;

namespace MultiDisplayWallpaperChanger {
  public partial class Form1 : Form {
    public Form1()
    {
      this.ResizeRedraw = true;

      InitializeComponent();

      SystemEvents.DisplaySettingsChanged += delegate {
        UpdateScreenBounds();
        DrawWallpaper();
      };

      checkBoxTile.CheckedChanged += delegate {
        DrawWallpaper();
      };

      UpdateScreenBounds();
    }

    private void UpdateScreenBounds()
    {
      allScreenBounds = new Rectangle(0, 0, 0, 0);

      foreach (var screen in Screen.AllScreens) {
        allScreenBounds = Rectangle.Union(allScreenBounds, screen.Bounds);
        screenBoxBoundsMap[screen.DeviceName] = Rectangle.Empty;
      }

      Refresh();
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      var g = e.Graphics;
      var scale = this.ClientSize.Width / (float)allScreenBounds.Width;

      var allScreenBoxBounds = new Rectangle();

      allScreenBoxBounds.X = 0;
      allScreenBoxBounds.Width = this.ClientSize.Width;
      allScreenBoxBounds.Height = (int)(allScreenBounds.Height * scale);
      allScreenBoxBounds.Y = label1.Bottom + (int)((this.ClientSize.Height - label1.Bottom) - allScreenBoxBounds.Height) / 2;

      g.Clear(SystemColors.Control);
      g.FillRectangle(SystemBrushes.AppWorkspace, allScreenBoxBounds);

      foreach (var screen in Screen.AllScreens) {
        var screenBoxBounds = new Rectangle();

        screenBoxBounds.X = allScreenBoxBounds.X + (int)((-allScreenBounds.X + screen.Bounds.X) * scale);
        screenBoxBounds.Y = allScreenBoxBounds.Y + (int)((-allScreenBounds.Y + screen.Bounds.Y) * scale);
        screenBoxBounds.Width  = (int)(screen.Bounds.Width  * scale);
        screenBoxBounds.Height = (int)(screen.Bounds.Height * scale);

        screenBoxBoundsMap[screen.DeviceName] = screenBoxBounds; // save calculated value

        if (wallpaperMap.ContainsKey(screen.DeviceName) && wallpaperMap[screen.DeviceName] != null)
          g.DrawImage(wallpaperMap[screen.DeviceName], screenBoxBounds);
        else
          g.FillRectangle(SystemBrushes.Window, screenBoxBounds);

        g.DrawRectangle(SystemPens.WindowFrame, screenBoxBounds);

        using (var sf = new StringFormat(StringFormat.GenericDefault)) {
          sf.Alignment = StringAlignment.Far;
          sf.LineAlignment = StringAlignment.Far;
          g.DrawString(screen.Bounds.ToString(), SystemFonts.DefaultFont, SystemBrushes.WindowText, screenBoxBounds, sf);
        }

        using (var f = new Font(SystemFonts.DefaultFont.FontFamily, screenBoxBounds.Height * 0.15f, GraphicsUnit.Pixel)) {
          using (var sf = new StringFormat(StringFormat.GenericDefault)) {
            sf.Alignment = StringAlignment.Center;
            sf.LineAlignment = StringAlignment.Center;
            g.DrawString(screen.DeviceName, f, SystemBrushes.WindowText, screenBoxBounds, sf);
          }
        }
      }

      base.OnPaint(e);
    }

    protected override void OnClick(EventArgs e)
    {
      try {
        Screen screen = null;

        foreach (var s in Screen.AllScreens) {
          if (screenBoxBoundsMap[s.DeviceName].Contains(PointToClient(Cursor.Position))) {
            screen = s;
            break;
          }
        }

        if (screen == null)
          return;

        using (var dialog = new OpenFileDialog()) {
          dialog.Filter = string.Format("{1} ({0})|{0}|{3} ({2})|{2}", "*.bmp;*.jpg;*.png", "画像ファイル", "*.*", "すべてのファイル");
          dialog.FilterIndex = 0;
          dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);

          if (dialog.ShowDialog() != DialogResult.OK)
            return;

          if (wallpaperMap.ContainsKey(screen.DeviceName) && wallpaperMap[screen.DeviceName] != null) {
            wallpaperMap[screen.DeviceName].Dispose();
            wallpaperMap[screen.DeviceName] = null;
          }

          try {
            wallpaperMap[screen.DeviceName] = new Bitmap(dialog.FileName);

            DrawWallpaper();
          }
          catch (Exception ex) {
            MessageBox.Show(this, string.Format("壁紙を設定できませんでした。{0}{1}", Environment.NewLine, ex));
          }
        } // using dialog
      }
      finally {
        base.OnClick(e);
      }
    }

    private void DrawWallpaper()
    {
      const string combinedWallpaperFile = @"wallpaper.bmp";

      using (var allScreenBitmap = new Bitmap(allScreenBounds.Width, allScreenBounds.Height, PixelFormat.Format24bppRgb)) {
        using (var g = Graphics.FromImage(allScreenBitmap)) {
          g.Clear(Color.Black);
          g.TranslateTransform((float)-allScreenBounds.X, (float)-allScreenBounds.Y, MatrixOrder.Prepend);

          foreach (var screen in Screen.AllScreens) {
            if (!wallpaperMap.ContainsKey(screen.DeviceName))
              continue;

            var bitmap = wallpaperMap[screen.DeviceName];

            if (bitmap == null)
              continue;

            if (checkBoxTile.Checked) {
              using (var clipRegion = new Region(screen.Bounds)) {
                g.Clip = clipRegion;

                for (var y = screen.Bounds.Top; y < screen.Bounds.Bottom; y += bitmap.Height) {
                  for (var x = screen.Bounds.Left; x < screen.Bounds.Right; x += bitmap.Width) {
                    g.DrawImageUnscaled(bitmap, x, y);
                  }
                }
              }
            }
            else {
              g.DrawImage(bitmap, screen.Bounds);
            }
          } // using g
        } // using allScreenBitmap

        allScreenBitmap.Save(combinedWallpaperFile, ImageFormat.Bmp);
      }

      SetWallpaper(combinedWallpaperFile, allScreenBounds.X, allScreenBounds.Y);
    }

    private void SetWallpaper(string file, int x, int y)
    {
      const int SPI_SETDESKWALLPAPER  = 0x0014;
      const int SPIF_UPDATEINIFILE    = 0x0001;
      const int SPIF_SENDWININICHANGE = 0x0002;

      using (var regkeyDesktop = Registry.CurrentUser.OpenSubKey(@"Control Panel\Desktop", true)) {
        if (regkeyDesktop == null)
          throw new Win32Exception(Marshal.GetLastWin32Error());

        regkeyDesktop.SetValue("TileWallpaper", "1");
        regkeyDesktop.SetValue("Wallpaperstyle", "0");
        regkeyDesktop.SetValue("WallpaperOriginX", x.ToString());
        regkeyDesktop.SetValue("WallpaperOriginY", y.ToString());
        regkeyDesktop.SetValue("Wallpaper", file);
      }

      var flags = (Environment.OSVersion.Platform == PlatformID.Win32NT)
        ? SPIF_SENDWININICHANGE
        : SPIF_SENDWININICHANGE | SPIF_UPDATEINIFILE;

      if (!SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, file, flags))
        throw new Win32Exception(Marshal.GetLastWin32Error());
    }

    [DllImport("user32.dll", SetLastError = true)] public static extern bool SystemParametersInfo(int uAction, int uParam, string lpvParam, int fuWinIini);

    private Rectangle allScreenBounds;
    private Dictionary<string, Rectangle> screenBoxBoundsMap = new Dictionary<string, Rectangle>();
    private Dictionary<string, Bitmap> wallpaperMap = new Dictionary<string, Bitmap>();
  }
}
